{
  open Printf
  open Parser
  module StringSet = Set.Make(String)
  exception UnexpectedToken of string
  type ident_buffer = { mutable name : string; }
  let buffer : ident_buffer = { name = "" }
  let temp : ident_buffer = { name = "" }
  let fmt : ident_buffer = { name = "" }

  type flags_type = { mutable in_print : bool; }
  let flags = { in_print = false; }

  let standardize name = String.map (fun x -> if x = '\\' then '_' else x) name
(*  type token =*)
(*    (* whitespaces  *)*)
(*      NL*)
(*    | TAB*)
(*    | SPACE*)
(*    (* identifier *)*)
(*    | ID of string*)
(*    (* literals *)*)
(*    | LIT_FL of float*)
(*    | LIT_INT of int*)
(*    | DIGIT of int*)
(*    (* separators *)*)
(*    | AMP*)
(*    | DBS*)
(*    | PCT*)
(*    | COMMA*)
(*    (* basics *)*)
(*    | VAR of char*)
(*    (* formatting *)*)
(*    | SUP*)
(*    | SUB*)
(*    (* enclosures *)*)
(*    | LPRN*)
(*    | RPRN*)
(*    | ABS*)
(*    | LBRC*)
(*    | RBRC*)
(*    | LFLOOR*)
(*    | RFLOOR*)
(*    | LCEIL*)
(*    | RCEIL*)
(*    (* envs *)*)
(*    | LCASE*)
(*    | RCASE*)
(*    (* operators *)*)
(*    | ADD*)
(*    | MINUS*)
(*    | MUL*)
(*    | DIV*)
(*    | EQ*)
(*    | LT*)
(*    | GT*)
(*    (* operators from command *)*)
(*    | LEQ*)
(*    | GEQ*)
(*    | NEQ*)
(*    | MID*)
(*    | NMID*)
(*    (*  logic  *)*)
(*    | NEG*)
(*    | WEDGE*)
(*    | VEE*)
(*    (* special functions *)*)
(*    | SUM*)
(*    | PROD*)
(*    | FRAC*)
(*    | BINOM*)
(*    (* binary functions *)*)
(*    | MAX*)
(*    | MIN*)
(*    | GCD*)
(*    (* basic functions *)*)
(*    | SQRT*)
(*    | EXP*)
(*    (* trigonometric functions *)*)
(*    | SIN*)
(*    | COS*)
(*    | TAN*)
(*    | COT*)
(*    | CSC*)
(*    | SEC*)
(*    | SINH*)
(*    | TANH*)
(*    | COTH*)
(*    | COSH*)
(*    | ARCSIN*)
(*    | ARCCOS*)
(*    | ARCTAN*)
(*    (* arithmetic functions *)*)
(*    | MOD*)
(*    (* log functions *)*)
(*    | LG*)
(*    | LN*)
(*    | LOG*)
(*    (*  bad  *)*)
(*    | BAD_CMD of string*)
(*    (*  misc  *)*)
(*    | CHAR of char*)
(*    | EOF*)
(*    | EOL*)
  let print_token t h =
    let _ = match t with
(*        NL -> if h then Printf.printf "\n" else Printf.printf "EOL"*)
(*      | TAB -> if h then Printf.printf "\t"  *)
(*      | SPACE -> if h then Printf.printf " "*)
      (* identifier *)
      | ID(s) -> if h then Printf.printf "ID(%s)" s else Printf.printf "ID"
      (* literals *)
      | LIT_FL(f) -> if h then Printf.printf "FL(%f)" f else Printf.printf "LIT_FL"
      | LIT_INT(i) -> if h then Printf.printf "INT(%d)" i else Printf.printf "LIT_INT"
      | DIGIT(i) -> if h then Printf.printf "DIGIT(%d)" i else Printf.printf "DIGIT"
      (* separators *)
      | AMP -> if h then Printf.printf "&" else Printf.printf "AMP"
      | DBS -> if h then Printf.printf "\n" else Printf.printf "DBS"
      | PCT -> if h then Printf.printf "%%" else Printf.printf "PCT"
      | COMMA -> if h then Printf.printf "," else Printf.printf "COMMA"
      (* basics *)
(*      | VAR(v) -> if h then Printf.printf "VAR(%c)" v*)
      (* formatting *)
      | SUP -> if h then Printf.printf "^" else Printf.printf "SUP"
      | SUB -> if h then Printf.printf "_" else Printf.printf "SUB"
      (* enclosures *)
      | LPRN -> if h then Printf.printf "(" else Printf.printf "LPRN"
      | RPRN -> if h then Printf.printf ")" else Printf.printf "RPRN"
      | LABS -> if h then Printf.printf "|" else Printf.printf "LABS"
      | RABS -> if h then Printf.printf "|" else Printf.printf "RABS"
      | LBRC -> if h then Printf.printf "{" else Printf.printf "LBRC"
      | RBRC -> if h then Printf.printf "}" else Printf.printf "RBRC"
      | LFLOOR -> Printf.printf "LFLOOR"
      | RFLOOR -> Printf.printf "RFLOOR"
      | LCEIL -> Printf.printf "LCEIL"
      | RCEIL -> Printf.printf "RCEIL"
      (* envs *)
      | LCASE -> if h then Printf.printf "\\begin{case}" else Printf.printf "LCASE"
      | RCASE -> if h then Printf.printf "\\end{case}" else Printf.printf "RCASE"
      | LSPLIT -> if h then Printf.printf "\\begin{split}" else Printf.printf "LSPLIT"
      | RSPLIT -> if h then Printf.printf "\\end{split}" else Printf.printf "RSPLIT"
      (* operators *)
      | ADD -> if h then Printf.printf "+" else Printf.printf "ADD"
      | MINUS -> if h then Printf.printf "-" else Printf.printf "MINUS"
      | MUL -> if h then Printf.printf "*" else Printf.printf "MUL"
      | DIV -> if h then Printf.printf "/" else Printf.printf "DIV"
      | EQ -> if h then Printf.printf "=" else Printf.printf "EQ"
      | LT -> if h then Printf.printf "<" else Printf.printf "LT"
      | GT -> if h then Printf.printf ">" else Printf.printf "GT"
      | LEQ -> if h then Printf.printf "≤" else Printf.printf "LEQ"
      | GEQ -> if h then Printf.printf "≥" else Printf.printf "GEQ"
      | NEQ -> if h then Printf.printf "≠" else Printf.printf "NEQ"
      | MID -> if h then Printf.printf "∣" else Printf.printf "MID"
      | NMID -> if h then Printf.printf "∤" else Printf.printf "NMID"
      (*  logic  *)
      | NEG -> if h then Printf.printf "¬" else Printf.printf "NEG"
      | WEDGE -> if h then Printf.printf "∧" else Printf.printf "WEDGE"
      | VEE -> if h then Printf.printf "∨" else Printf.printf "VEE"
      (* special functions *)
      | SUM -> if h then Printf.printf "∑" else Printf.printf "SUM"
      | PROD -> if h then Printf.printf "∏" else Printf.printf "PROD"
      | FRAC -> Printf.printf "FRAC"
      | BINOM -> Printf.printf "BINOM"
      (* binary functions *)
      | MAX ->  Printf.printf "MAX"
      | MIN -> Printf.printf "MIN"
      | GCD -> Printf.printf "GCD"
      (* basic functions *)
      | SQRT -> if h then Printf.printf "√" else Printf.printf "SQRT"
(*      | EXP -> Printf.printf "EXP"*)
      (* trigonometric functions *)
      | SIN -> Printf.printf "SIN"
      | COS -> Printf.printf "COS"
      | TAN -> Printf.printf "TAN"
      | COT -> Printf.printf "COT"
      | CSC -> Printf.printf "CSC"
      | SEC -> Printf.printf "SEC"
      | SINH -> Printf.printf "SINH"
      | TANH -> Printf.printf "TANH"
      | COTH -> Printf.printf "COTH"
      | COSH -> Printf.printf "COSH"
      | ARCSIN -> Printf.printf "ARCSIN"
      | ARCCOS -> Printf.printf "ARCCOS"
      | ARCTAN -> Printf.printf "ARCTAN"
      (* arithmetic functions *)
      | MOD -> if h then Printf.printf "(mod)" else Printf.printf "MOD"
      (* log functions *)
      | LG -> Printf.printf "LG"
      | LN -> Printf.printf "LN"
      | LOG -> Printf.printf "LOG"
      (*  bad  *)
(*      | BAD_CMD(s) -> if h then Printf.printf "BAD_CMD(%s)" s*)
      (*  misc  *)
(*      | CHAR(c) -> if h then Printf.printf "CHAR(%c)" c*)
      | EOL -> Printf.printf "EOL"
      | EOF -> Printf.printf "EOF"

      | _ -> Printf.printf "Unimplemented"
    in let () = Printf.printf " "
    in ()

  exception InvalidCommand of string
  let valid_greek_letters = StringSet.of_list [
    "alpha" ;
    "beta" ;
    "Gamma" ;
    "gamma" ;
    "Delta" ;
    "delta" ;
    "epsilon" ;
    "varepsilon" ;
    "zeta" ;
    "eta" ;
    "Theta" ;
    "theta" ;
    "vartheta" ;
    "iota" ;
    "kappa" ;
    "varkappa" ;
    "Lambda";
    "lambda" ;
    "mu" ;
    "nu" ;
    "Xi" ;
    "xi" ;
    "Pi" ;
    "pi" ;
    "varpi" ;
    "rho" ;
    "varrho" ;
    "Sigma" ;
    "sigma" ;
    "varsigma" ;
    "tau" ;
    "Upsilon" ;
    "upsilon" ;
    "Phi" ;
    "phi" ;
    "varphi" ;
    "chi" ;
    "Psi" ;
    "psi" ;
    "Omega" ;
    "omega" ;
  ]

  let parse_greek raw = let letter = String.sub raw 1 ((String.length raw) - 1)
    in match StringSet.find_opt letter valid_greek_letters with
      | Some(l) -> Some(letter)
      | None -> None

  let parse_greek_in_ident raw = match parse_greek raw with
    | Some(l) -> buffer.name <- buffer.name ^ "__" ^ l
    | None -> raise (InvalidCommand raw)

  let parse_command raw = let command = String.sub raw 1 ((String.length raw) - 1)
    in match command with
      | "times" -> MUL
      | "cdot" -> MUL
      | "div" -> DIV
      | "leq" -> LEQ
      | "geq" -> GEQ
      | "neq" -> NEQ
      | "mid" -> MID
      | "nmid" -> NMID
      (*  logic  *)
      | "neg" -> NEG
      | "wedge" -> WEDGE
      | "vee" -> VEE
      (* special functions *)
      | "sum" -> SUM
      | "prod" -> PROD
      | "frac" -> FRAC
      | "binom" -> BINOM
      (* binary functions *)
      | "max" -> MAX
      | "min" -> MIN
      | "gcd" -> GCD
      (* basic functions *)
      | "sqrt" -> SQRT
(*      | "exp" -> EXP*)
      (* trigonometric functions *)
      | "sin" -> SIN
      | "cos" -> COS
      | "tan" -> TAN
      | "cot" -> COT
      | "sec" -> SEC
      | "csc" -> CSC
      | "sinh" -> SINH
      | "cosh" -> COSH
      | "tanh" -> TANH
      | "csc" -> CSC
      | "coth" -> COTH
      | "arcsin" -> ARCSIN
      | "arccos" -> ARCCOS
      | "arctan" -> ARCTAN
      (* arithmetic functions *)
      | "mod" -> MOD
      (* log functions *)
      | "lg" -> LG
      | "ln" -> LN
      | "log" -> LOG
      | _ -> match parse_greek raw with
        | Some(l) -> ID("__" ^ l)
        | None -> raise (UnexpectedToken raw)
}

let cmd = '\\'['a'-'z' 'A'-'Z']+
let int_r = ['0'-'9'] ['0'-'9']+
let int_part = ['0'-'9']+
let float_r = (int_part? '.' int_part) | (int_part '.')
let formatter =
    "mathrm"
  | "mathit"
  | "mathbf"
  | "mathsf"
  | "mathtt"
  | "mathfrak"
  | "mathcal"
  | "mathbb"
  | "mathscr"

rule ctex_read_token = parse
  (* whitespaces  *)
  | [' ' '\t' '\r'] { ctex_read_token lexbuf }
  | ['\n'] { if flags.in_print then let () = flags.in_print <- false in EOL else ctex_read_token lexbuf }
  (* separators *)
  | "\\\\" { DBS }
  | '&' { AMP }
  | "%%" { ctex_read_comment lexbuf; if flags.in_print then let () = flags.in_print <- false in EOL else ctex_read_token lexbuf }
  | '%' { let () = flags.in_print <- true in PCT }
  | ',' { COMMA }
  (* envs *)
  | "\\begin{cases}" { LCASE }
  | "\\end{cases}" { RCASE }
  | "\\begin{split}" { LSPLIT }
  | "\\end{split}" { RSPLIT }
  (* enclosures *)
  | "(" { LPRN }
  | ")" { RPRN }
  | "\\left\\|" { LABS }
  | "\\right\\|" { RABS }
  | "{" { LBRC }
  | "}" { RBRC }
  | "\\lfloor" { LFLOOR }
  | "\\rfloor" { RFLOOR }
  | "\\lceil" { LCEIL }
  | "\\rceil" { RCEIL }
  (* formatting *)
  | '^' { SUP }
  | '_' { SUB }
  (* operatings *)
  | '+' { ADD }
  | '-' { MINUS }
  | '*' { MUL }
  | '/' { DIV }
  | '=' { EQ }
  | '<' { LT }
  | '>' { GT }
  (* literals *)
  | float_r as float_s { LIT_FL(float_of_string float_s) }
  | int_r as int_s { LIT_INT(int_of_string int_s) }
  | ['0'-'9'] as d_s { DIGIT(int_of_string (String.make 1 d_s)) }
  (* complex identifier *)
  | '\\' formatter as f '{' { fmt.name <- f ; ctex_read_ident lexbuf; ID(standardize(fmt.name ^ "___"^ String.trim temp.name ^ "___")) }
  (* commands, including functions and part of enclosures, IDs, operators *)
  | cmd as s { parse_command s }
  (* identifier *)
  | ['a'-'z' 'A'-'Z'] as v { ID(String.make 1 v) }
  (* misc *)
  | _ as c { raise (UnexpectedToken (String.make 1 c)) }
  | eof { EOF }
and ctex_read_comment = parse
  | '\n' | eof { () }
  | _ { ctex_read_comment lexbuf }
and ctex_read_ident = parse
  | [' ' '\t' '\r' '\n'] { ctex_read_ident lexbuf }
  | '}' { temp.name <- buffer.name ; buffer.name <- "" ; () }
  | cmd as s { parse_greek_in_ident s ; ctex_read_ident lexbuf }
  | ['a'-'z' 'A'-'Z'] as c { buffer.name <- buffer.name ^ String.make 1 c ; ctex_read_ident lexbuf }
  | _ as c { raise (UnexpectedToken (String.make 1 c)) }
